// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package irc

import (
	"fmt"
	"strings"
)

type Message struct {
	Command string
	Raw     string
	Nick    string
	Host    string
	Source  string
	User    string
	Args    []string
}

func (i *Context) Pong(host string) {
	msg := Message{
		Command: "PONG",
		Args:    []string{":" + host},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Pass() {
	msg := Message{
		Command: "PASS",
		Args:    []string{i.network.pass},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Nick() {
	msg := Message{
		Command: "NICK",
		Args:    []string{i.network.nick},
	}
	i.outgoingMsg <- msg
}

func (i *Context) User() {
	msg := Message{
		Command: "USER",
		Args: []string{
			i.network.user,
			"* 8",
			":" + i.network.realName,
		},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Join(channel string, key string) {
	msg := Message{
		Command: "JOIN",
		Args:    []string{channel, key},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Part(channel string, text string) {
	msg := Message{
		Command: "PART",
		Args: []string{
			channel,
			":" + text,
		},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Quit(text string) {
	msg := Message{
		Command: "QUIT",
		Args:    []string{":" + text},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Privmsg(channel string, text string) {
	msg := Message{
		Command: "PRIVMSG",
		Args: []string{
			channel,
			":" + text,
		},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Notice(nick string, text string) {
	msg := Message{
		Command: "NOTICE",
		Args: []string{
			nick,
			":" + text,
		},
	}
	i.outgoingMsg <- msg
}

func (i *Context) Topic(channel string, text string) {
	msg := Message{
		Command: "TOPIC",
		Args: []string{
			channel,
			":" + text,
		},
	}
	i.outgoingMsg <- msg
}

func (i *Context) ParseRawMessage(raw string) (*Message, error) {
	var msg Message
	msg.Raw = raw
	raw = strings.TrimSpace(raw)
	if raw[0] == ':' {
		if i := strings.Index(raw, " "); i != -1 {
			msg.Source = raw[1:i]
			raw = raw[1:]
		} else {
			return nil, fmt.Errorf("Could not parse source, ignoring...")
		}
		if i := strings.Index(raw, "!"); i != -1 {
			msg.Nick = raw[:i]
			raw = raw[i+1:]
		} else {
			return nil, fmt.Errorf("Could not parse nick, ignoring...")
		}
		if i := strings.Index(raw, "@"); i != -1 {
			msg.User = raw[:i]
			raw = raw[i+1:]
		} else {
			return nil, fmt.Errorf("Could not parse user, ignoring...")
		}
		if i := strings.Index(raw, " "); i != -1 {
			msg.Host = raw[:i]
			raw = raw[i+1:]
		} else {
			return nil, fmt.Errorf("Could not parse host, ignoring...")
		}
	}
	if i := strings.Index(raw, " "); i != -1 {
		msg.Command = raw[:i]
		raw = raw[i+1:]
	} else {
		return nil, fmt.Errorf("Could not parse command, ignoring...")
	}
	if raw[0] == ':' {
		msg.Args = []string{raw[1:]}
	} else {
		msg.Args = strings.SplitN(raw, " :", 2)
	}
	return &msg, nil
}

// Unpack a message into a byte array
func (i *Context) UnpackMessage(msg Message) []byte {
	rawMsg := msg.Command + " " + strings.Join(msg.Args, " ")
	rawMsg = strings.TrimSpace(rawMsg) + "\r\n"
	return []byte(rawMsg)
}

func (i *Context) incomingMsgLoop() error {
	for {
		select {
		case msg, ok := <-i.incomingMsg:
			if !ok {
				return nil
			}
			i.dispatchHandler(msg)
		}
	}
	return nil
}

func (i *Context) outgoingMsgLoop() error {
	for {
		select {
		case msg, ok := <-i.outgoingMsg:
			if !ok {
				return nil
			}
			rawMsg := i.UnpackMessage(msg)
			err := i.TxRawMessage(rawMsg)
			if err != nil {
				return err
			}
		}
	}
	return nil
}
