// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package mapfs

import (
	"bufio"
	"bytes"
	"code.google.com/p/go.crypto/blowfish"
	"encoding/gob"
	"fmt"
	"io"
	"log"
	"os"
)

type BlowfishMap struct {
	cache  map[int]blowfishVal
	cipher *blowfish.Cipher
}

type blowfishVal struct {
	dirty bool
	buf   []byte // Actual encrypted data
	siz   int    // Length of decrypted data
}

type hdr struct {
	Magic string
	Len   int
	Data  []byte
}

func NewBlowfishMap(key string) *BlowfishMap {
	c, err := blowfish.NewCipher([]byte(key))
	if err != nil {
		log.Fatal(err)
	}
	return &BlowfishMap{
		cache:  make(map[int]blowfishVal),
		cipher: c,
	}
}

func (b *BlowfishMap) rawRead(path string) ([]byte, error) {
	fi, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer fi.Close()
	r := bufio.NewReader(fi)
	g := gob.NewDecoder(r)
	h := new(hdr)
	err = g.Decode(h)
	if err != nil {
		if err != io.EOF {
			return nil, err
		}
	}
	pt := b.decryptBuf(h.Data)
	return pt[0:h.Len], nil
}

func (b *BlowfishMap) rawWrite(path string, buf []byte) error {
	fo, err := os.Create(path)
	if err != nil {
		return err
	}
	defer fo.Close()
	ct := b.encryptBuf(buf)
	w := bufio.NewWriter(fo)
	h := &hdr{"BENC", len(buf), ct}
	g := gob.NewEncoder(w)
	err = g.Encode(*h)
	if err != nil {
		return err
	}
	err = w.Flush()
	if err != nil {
		return err
	}
	return nil
}

func (b *BlowfishMap) get(key int) ([]byte, error) {
	val, ok := b.cache[key]
	if !ok {
		return nil, fmt.Errorf("No entry with key: %d", key)
	}
	return val.buf, nil
}

func (b *BlowfishMap) put(key int, buf []byte) error {
	_, ok := b.cache[key]
	if ok {
		return fmt.Errorf("Key %d already in use", key)
	}
	b.cache[key] = blowfishVal{true, buf, len(buf)}
	return nil
}

func (b *BlowfishMap) add(buf []byte) (int, error) {
	key := len(b.cache)
	b.cache[key] = blowfishVal{true, buf, len(buf)}
	return key, nil
}

func (b *BlowfishMap) countMatches(buf []byte) int {
	i := 0
	for _, v := range b.cache {
		if bytes.Equal(v.buf, buf) {
			i++
		}
	}
	return i
}

func (b *BlowfishMap) virtSync() error {
	for k, v := range b.cache {
		b.cache[k] = blowfishVal{false, v.buf, v.siz}
	}
	return nil
}

func (b *BlowfishMap) syncEntry(path string, key int) error {
	val, ok := b.cache[key]
	if !ok {
		return fmt.Errorf("No entry with key: %d", key)
	}
	if val.dirty {
		b.rawWrite(path, val.buf)
	}
	return nil
}

func (b *BlowfishMap) size() int {
	return len(b.cache)
}

func (b *BlowfishMap) string() string {
	s := ""
	for k, v := range b.cache {
		dirty := ""
		if v.dirty {
			dirty = "yes"
		} else {
			dirty = "no"
		}
		s += fmt.Sprintf("k: %d - dirty: %s\n",
			k, dirty)
	}
	return s
}

func (b *BlowfishMap) encryptBuf(a []byte) []byte {
	padded := a
	diff := len(a) % blowfish.BlockSize
	if diff != 0 {
		diff = blowfish.BlockSize - diff
	}
	for i := 0; i < diff; i++ {
		padded = append(padded, byte(0x0))
	}
	ct := make([]byte, len(padded))
	for i := 0; i < len(a); i += blowfish.BlockSize {
		if i+blowfish.BlockSize > len(a) {
			b.cipher.Encrypt(ct[i:], padded[i:])
			return ct
		}
		b.cipher.Encrypt(ct[i:i+blowfish.BlockSize],
			padded[i:i+blowfish.BlockSize])
	}
	return ct
}

func (b *BlowfishMap) decryptBuf(a []byte) []byte {
	pt := make([]byte, len(a))
	for i := 0; i < len(a); i += blowfish.BlockSize {
		b.cipher.Decrypt(pt[i:i+blowfish.BlockSize],
			a[i:i+blowfish.BlockSize])
	}
	return pt
}
