// Copyright 2013 TLH and dsp. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package mapfs

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
)

type PlainMap struct {
	cache map[int]plainVal
}

type plainVal struct {
	dirty bool
	buf   []byte // Actual raw data
	siz   int    // Size of raw data
}

func NewPlainMap() *PlainMap {
	return &PlainMap{
		cache: make(map[int]plainVal),
	}
}

func (p *PlainMap) rawRead(path string) ([]byte, error) {
	b, err := ioutil.ReadFile(path)
	if err != nil {
		return nil, err
	}
	return b, err
}

func (p *PlainMap) rawWrite(path string, buf []byte) error {
	_, err := os.Stat(path)
	if err == nil {
		return fmt.Errorf("File %s already exists", path)
	}
	err = ioutil.WriteFile(path, buf, 0644)
	if err != nil {
		return err
	}
	return nil
}

func (p *PlainMap) get(key int) ([]byte, error) {
	val, ok := p.cache[key]
	if !ok {
		return nil, fmt.Errorf("No entry with key: %d", key)
	}
	return val.buf, nil
}

func (p *PlainMap) put(key int, buf []byte) error {
	_, ok := p.cache[key]
	if ok {
		return fmt.Errorf("Key %d already in use", key)
	}
	p.cache[key] = plainVal{true, buf, len(buf)}
	return nil
}

func (p *PlainMap) add(buf []byte) (int, error) {
	key := len(p.cache)
	p.cache[key] = plainVal{true, buf, len(buf)}
	return key, nil
}

func (p *PlainMap) countMatches(buf []byte) int {
	i := 0
	for _, v := range p.cache {
		if bytes.Equal(v.buf, buf) {
			i++
		}
	}
	return i
}

func (p *PlainMap) virtSync() error {
	for k, v := range p.cache {
		p.cache[k] = plainVal{false, v.buf, v.siz}
	}
	return nil
}

func (p *PlainMap) syncEntry(path string, key int) error {
	val, ok := p.cache[key]
	if !ok {
		return fmt.Errorf("No entry with key: %d", key)
	}
	if val.dirty {
		p.rawWrite(path, val.buf)
	}
	return nil
}

func (p *PlainMap) size() int {
	return len(p.cache)
}

func (p *PlainMap) string() string {
	s := ""
	for k, v := range p.cache {
		dirty := ""
		if v.dirty {
			dirty = "yes"
		} else {
			dirty = "no"
		}
		s += fmt.Sprintf("k: %d - dirty: %s\n",
			k, dirty)
	}
	return s
}
